/* Copyright (c) 2006, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

//const char *idx_fname = "3dsounds.idx";
//const char *snd_fname = "3dsounds.wav";
//const char *writeto = "sounds";

void findspace(char *buf, int *eger);

int main(int argc, char *argv[]) {
  
  char *input;
  if (argc != 2) {
    input = malloc(1024);
    printf("Read which file? (You can put a pathname here, too, if you need to)\n> ");
    fflush(stdout);
    fgets(input, 1024, stdin);
    int i;
    for (i = 0; input[i] != 10 && input[i] != 13 && input[i] != 0; i++);
    input[i] = 0;
  } else {
    input = argv[1];
  }
  
  int l = strlen(input);
  int fpos = -1; // Position of actual filename (path ignored)
  int eger;
  for (eger = 0; eger < l; eger++)
    if (input[eger] == '/' || input[eger] == '\\')
      fpos = eger;
  fpos++; // So it doesn't point to the last slash, but to the file itself.
  char *writeto = malloc(l + 5 - fpos);
  char *idx_fname = malloc(l + 5);
  char *snd_fname = malloc(l + 5);
  sprintf(idx_fname, "%s.idx", input);
  sprintf(snd_fname, "%s.wav", input);
  memcpy(writeto,input + fpos, l - fpos);
  
  FILE *idx = fopen(idx_fname, "r");
  if (!idx) {
    printf("Unable to open %s for reading!\n",idx_fname);
    exit(1);
  }
  FILE *snd = fopen(snd_fname, "r");
  if (!snd) {
    printf("Unable to open %s for reading!\n",snd_fname);
    exit(2);
  }
  
  fseek(snd, 0, SEEK_END);
  int sndsize = ftell(snd);

  char *buf = malloc(2048);
  
  // Index (IDX) lists the following information for every Planetside sound:
  char *fname = malloc(2048); // Filename
  int offset = 0; // Offset in 3dsounds.wav file
  int size = 0; // Size of individual wav
  int srate = 0; // Sampling frequency/rate
  int bitdepth = 0; // Bit depth

  char *tmp = malloc(1024); // So we can atoi() numbers.
  
  // Oddly, there are only 3 sounds that're at 11025Hz/8bit, the rest are 22050/16bit...



  // Make the directory first.  Rather than read the sound index into memory, we read it as we go, so we'll do this before reading in the index itself.

  printf("Extracting files to %s/...\n",writeto);
  mkdir(writeto, S_IRUSR | S_IWUSR | S_IXUSR | S_IXGRP | S_IXOTH | S_IRGRP | S_IROTH);
  if (chdir(writeto) != 0) {
    printf("Couldn't change to %s/!\n",writeto);
    exit(3);
  }
  
  int maxsize = 0;

  int lastwaseight = 0; // For some bug regarding 8-bit sounds.
  
  while (fgets(buf,2048,idx)) {
    int start = 0;
    int end = 0;
    findspace(buf, &end);
    memcpy(fname, buf + start, end - start);
    fname[end - start] = 0;
    
    start = ++end; // Start seeking from next position, skipping that space.
    findspace(buf,&end);
    memcpy(tmp,buf + start, end - start);
    tmp[end - start] = 0;
    offset = atoi(tmp);

    start = ++end; // Start seeking from next position, skipping that space.
    findspace(buf,&end);
    memcpy(tmp,buf + start, end - start);
    tmp[end - start] = 0;
    size = atoi(tmp);
    if (size > maxsize)
      maxsize = size;

    start = ++end; // Start seeking from next position, skipping that space.
    findspace(buf,&end);
    memcpy(tmp,buf + start, end - start);
    tmp[end - start] = 0;
    srate = atoi(tmp);

    start = ++end; // Start seeking from next position, skipping that space.
    findspace(buf,&end);
    memcpy(tmp,buf + start, end - start);
    tmp[end - start] = 0;
    bitdepth = atoi(tmp);
    
    printf("% 4.1f%%, %s (%.3fMB): %iHz %ibit @ %08X\n",(float)offset/(float)sndsize*100, fname,(float)size/1024/1024, srate, bitdepth, offset);

    // Now, open up that file and write it out.
    
    if (offset == 0) {
      // Pull out that not-completely true RIFF/WAV header we have.
      offset += 44;
      size -= 44;
    }
    
    if (bitdepth == 8 || lastwaseight) {
      // Some useless data is at the beginning of all 8-bit sounds or sounds directly after an 8-bit sound.
      offset += 88;
      size -= 88;
      if (bitdepth == 8) {
	lastwaseight = 1;
      } else {
	lastwaseight = 0;
      }
    }
    FILE *out = fopen(fname, "w");
    unsigned char *wav = malloc(size);
    if (out && wav) {
      
      // First, let's write a WAV header for our file.
      unsigned char *riff = malloc(12);
      unsigned char *fmt = malloc(24);
      unsigned char *data = malloc(8);
      
      u_int32_t asdf;
      u_int16_t fdsa;
      // WAVE deals with 32bit ints, so let's preferably work with one, too.
      
	
      memcpy(riff, "RIFF", 4);
      asdf =
      4/*RIFF - id/size*/ +
	24/*FMT*/ + 
	8/*data*/ +
	size /*wave*/;
      
      memcpy(riff + 4, &asdf, 4); // Size of chunk (minus first 8 bytes)
      memcpy(riff + 8, "WAVE", 4);
      
      memcpy(fmt, "fmt ", 4);
      asdf = 16;
      memcpy(fmt + 4, &asdf, 4); // Size of chunk, again minus first 8
      fdsa = 1;
      memcpy(fmt + 8, &fdsa, 2); // AudioFormat (1=PCM)
      fdsa = 1;
      memcpy(fmt + 10, &fdsa, 2); // NumChannels (always mono)
      asdf = srate;
      memcpy(fmt + 12, &asdf, 4); // SampleRate
      asdf = srate * bitdepth / 8;
      memcpy(fmt + 16, &asdf, 4); // ByteRate (number of bytes in each second)
      fdsa = bitdepth / 8;
      memcpy(fmt + 20, &fdsa, 2); // BlockAlign (number of bytes in each smple)
      fdsa = bitdepth;
      memcpy(fmt + 22, &fdsa, 2); // BitsPerSample
      
      memcpy(data,"data", 4);
      asdf = size;
      memcpy(data + 4, &size, 4);
      
      
      fseek(snd, offset, SEEK_SET);
      int i;
      for (i = 0; i < size; i++)
	wav[i] = getc(snd);
      fwrite(riff, 12, 1, out);
      fwrite(fmt, 24, 1, out);
      fwrite(data, 8, 1, out);
      fwrite(wav, size, 1, out);
      fclose(out);
    } else if (!out) {
      printf("Unable to open %s for writing!\n", fname);
    } else if (!wav) {
      printf("Not enough memory to read %s into memory!\n",fname);
    }
    
    
  }
  exit(0);
}

void findspace(char *buf, int *eger) {
  while (buf[*eger] != 32 &&
	 buf[*eger] != 10 &&
	 buf[*eger] != 13 &&
	 buf[*eger] != 0 &&
	 buf[*eger] != -1) {
    (*eger)++;
  }
  return;
}


